%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Eotvos Lorand University.
%%% Portions created  by Eotvos  Lorand University are  Copyright 2007-2009,
%%% Eotvos Lorand University. All Rights Reserved.

%%% @doc This module gives the gen_fileman behaviour which is needed to implement
%%% file manager modules for different positioning methods.
%%% The core of this behaviour is a supervisor implemented in this module which
%%% starts the proper refcore_fileman_name ("name" is the name of a positioning
%%% method) module according to the current positioning mode chosen by the user.
%%%
%%% @author Gabor Hosszu

-module(gen_fileman).
         
         
-export([behaviour_info/1]).
         
-export([start_link/0,init/1]).
-export([restart/1]).

-export([add_file/1, add_file/2, drop_file/1, drop_file/2, save_file/1]).
-export([add_form/3, add_text/3, drop_form/2]).
-export([form_hash/1,create_scanner/0, progress_start/6, progress_step/1]).
-export([set_file_hash/2, file_hash/1]).

-export([hold_prefix/3,input_tokens/1,store_tokens/1,tokenize/1]).
-export([output_env/0,make_backup/1,real_forms/1,orig_text/2,form_hash_neq/2]).
-export([file_text/1,set_form_hash/2,create_file_node/2]).
-export([file_status/1,dep_forms/1,parse_form/4,token_length/1,split_forms/1]).
-export([split_stop_tokens/1,merge/2,includers/1,action_length/1,act_count/1]).
-export([update_lastmod/1,merge_ws/1, graph_update/2, add_new_file/2]).



-include("core.hrl").

-behaviour(supervisor).

-define(EnvKey, refcore_fileman_env).
-record(fmenv, {scanner}).

%%% ===========================================================================
%%% Supervisor callbacks

%% @private
init(PosModule) ->
    {ok,
     {{one_for_all, 3, 1},
      [
       {PosModule,
        {PosModule, start_link, []},
        permanent,
        ?SHUTDOWN,
        worker,
        [PosModule]}
      ]}}.

%%% ===========================================================================
%%% Behaviour core

behaviour_info(callbacks) ->
    [{add_file, 1}, {add_file, 2}, {drop_file, 1}, {drop_file, 2}, {save_file, 1}, 
        {add_form, 3}, {add_text, 3}, {drop_form, 2}, {start_link, 0},      
        {init,1},{handle_call, 3}, {handle_cast, 2}, {handle_info, 2},{terminate, 2}, {code_change, 3}].

%% @doc Starts supervisor for current posmode set by the user
start_link() ->
    PosMod = get_init_arg(pos),
    PosModule = create_modname(PosMod),
    ets:new(?PosModTab,[public,named_table]),
    ets:insert(?PosModTab,{posmod,PosMod}),
    supervisor:start_link({local, fileman_sup},?MODULE,PosModule).

%% @doc Restarts supervisor with the given posmode
%% The database must be reseted before this.
restart(PosMod) ->
    Chln = supervisor:which_children(fileman_sup),
    [ supervisor:terminate_child(fileman_sup,Id) ||
            {Id,_Pid,_,_} <- Chln ],
    [ supervisor:delete_child(fileman_sup,Id) ||
            {Id,_Pid,_,_} <- Chln ],
    ets:delete(?PosModTab,posmod),
    PModule = create_modname(PosMod),
    ?Syn:set_env(db_posmode, PosMod),
    ets:insert(?PosModTab,{posmod,PosMod}),
    supervisor:start_child(fileman_sup, 
        {PModule,
        {PModule, start_link, []},
        permanent,
        infinity,
        worker,
        [PModule]}).
%%  Pos argument initialization

get_init_arg(pos) ->
    ActPosMod =
    try 
        begin
            {ok, [[AStr]]} = init:get_argument(pos),
            {ok, Tokens, _} = erl_scan:string(AStr ++ "."),
            {ok, Term} = erl_parse:parse_term(Tokens),
            Term
        end 
    of
        X -> X
    catch
        _:_ -> undefined %% not needed
    end,
    DbLastPosMod =
    case ?Syn:get_env(db_posmode) of %TODO macro for this env
        [] -> undefined;
        [H] -> H;
        [_H | _ ] -> inconsistent
    end,
    init_posmode(DbLastPosMod, ActPosMod).

%%  Gives a posmode when last used db posmode and actual posmode argument are given
%%      sets the proper env node in db if change needed

init_posmode(undefined, A) -> 
    ?Syn:set_env(db_posmode, A),
    A;
init_posmode(inconsistent, A) -> 
    ?Syn:set_env(db_posmode, A),
    A; %% TODO: error reporting
init_posmode(DbLastPosMod, A) -> 
    case ?Query:exec([file]) of
        [] -> ?Syn:set_env(db_posmode, A),
                A;
        _ -> DbLastPosMod
    end.

handle_call({Fun,Args}) ->
    Mod = ets:lookup_element(?PosModTab,posmod,2),
    Module = create_modname(Mod),
    apply(Module,Fun,Args).

create_modname(PosMod) ->
    list_to_atom(lists:concat([refcore_fileman_,PosMod])).

%%% ===========================================================================
%%% Fileman interface

add_file(File) ->
    handle_call({add_file,[File]}).

add_file(A,B) ->
    handle_call({add_file,[A,B]}).
 
drop_file(File) ->
    handle_call({drop_file,[File]}).

drop_file(A,B) ->
    handle_call({drop_file,[A,B]}).
 
save_file(File) ->
    handle_call({save_file,[File]}).

add_form(A,B,C) ->
    handle_call({add_form,[A,B,C]}).

add_text(A,B,C) ->
    handle_call({add_text,[A,B,C]}).

drop_form(A,B) ->
    handle_call({drop_form,[A,B]}).


create_scanner() ->
    handle_call({create_scanner,[]}).

make_token(A,B,C,D) ->
    handle_call({make_token,[A,B,C,D]}).

graph_update(A,B) ->
    handle_call({graph_update,[A,B]}).

add_new_file(A,B) ->
    handle_call({add_new_file,[A,B]}).
    
    
improve_msg(File, Line, FormIndex, Tokens, Msg) ->
    handle_call({improve_msg,[File, Line, FormIndex, Tokens, Msg]}).


%%% ============================================================================
%%% Progress reporter

-record(progress, {op, starttime, file, max, lengths, count=1, totallen}).

progress_start(Op, Now, Path, Lengths, TotalLength, Max) ->
    #progress{op=Op, starttime = Now, file=Path, lengths=Lengths, totallen=TotalLength, max=Max}.

progress_step(Progress) -> progress_step(Progress, 1).

progress_step(Progress=#progress{op=Op, starttime=Start, count=Count, file=File, 
                                 lengths=[Length|RestLs], totallen=TLen, max=Max}, 
              N)
  when is_function(Op, 5) ->
    KBps =
        case sec_diff(Start, os:timestamp()) of
            0.0   -> 0;
%            0   -> 0;
            Sec -> Length / 1024 / Sec
        end,
    Percent =
        case {Max, TLen} of
            {0, _} -> 1;
            {_, 0} -> 1;
            _      -> Length / TLen
        end,
    Op(File, Percent, Count, Max, KBps),
    Progress#progress{lengths=RestLs, count=Count+N};
progress_step(Progress=#progress{op=none}, _) ->
    Progress.

%% todo Move to ?MISC.
sec_diff({MgS1, S1, MiS1}, {MgS2, S2, MiS2}) ->
    (MgS2 - MgS1) * 1000000 + S2 - S1 + (MiS2 - MiS1) / 1000000.



%%% ============================================================================
%%% Syntax level manipulation


%% @spec parse_form(File :: node(), integer(), module | header, referl_preproc:processedForm()) -> node()
parse_form(_File, _Index, module, {form, Form}) -> Form;
parse_form(File, Index, module, {tokens, Tokens}) -> parse(File, Index, Tokens);
parse_form(File, Index, module, {vtokens, Dep, Orig, Tokens}) ->
    Form = parse(File, Index, Tokens),
    ?Graph:mklink(Form, forig, Orig),
    ?Graph:mklink(Form, fdep, Dep),
    Data = ?ESG:data(Form),
    ?ESG:update(Form, Data#form{hash=virtual}),
    Form;
parse_form(_File, _Index, header, {form, Form}) -> Form;
parse_form(File, Index, header, {tokens, [{#token{type='-'}, _},
                             {#token{type=record},_}|_]=Tokens}) ->
    parse(File, Index, Tokens);
parse_form(_File, _Index, header, {tokens, Tokens}) ->
    Form = ?ESG:create(#form{type=lex, tag=store}),
    [?ESG:insert(Form, flex, Token) || {_, Token} <- Tokens],
    Form.


%% @spec parse(File :: node(), integer(), [{#token{}, node()}]) -> node()
%%
%% @doc Parses `Tokens' and returns the result, the root form of the tree.
parse(File, FormIndex, Tokens) ->
    TokenData =
        [{Type, 1, Data}
            || Data={#token{type = Type}, _} <- Tokens],
    case ?ErlParser:parse(TokenData) of
        {error, {Line, _Mod, Msg}} ->
            ErrMsg = improve_msg(File, Line, FormIndex, Tokens, Msg),
            ErrMsg == "" orelse
                error_logger:info_msg(ErrMsg),

            Form = ?ESG:create(#form{type=error, tag={1, ErrMsg}}),
            [begin
                 [?ESG:remove(P, T, N) || {T, P} <- ?Syn:parent(N)], %% possibly an unparsable expr
                 ?ESG:insert(Form, flex, N)
             end || {_, N} <- Tokens],
            Form;
        {ok, Result} -> Result
    end.

    
%%% ============================================================================
%%% File level manipulations

%% @doc Initializes a new file node for the loaded file.
create_file_node(Path, Eol) ->
    Type = case filename:extension(Path) of
               ".erl" -> module;
               _      -> header
           end,
    File = ?ESG:create(#file{type=Type, path=Path, eol=Eol, lastmod=now()}),
    ?ESG:insert(?ESG:root(), file, File),
    ?Graph:mklink(File, incl, File),
    File.

%% private type
%% eol() = {Style, Eof}
%%         Style = {'cr' | 'lf' | 'crlf'}
%%         Eof   = {'eol' | 'noeol'}

%% @spec file_text(string()) -> {string(), eol()}
%% @throws {error, string()}
%% @doc Reads the contents of a file, and returns it in a canonical format.
file_text(Path) ->
    %% TODO: move abs_path to referl_gen
    case file:read_file(?File:abs_path(Path)) of
        {ok, BinaryText} -> file_text(BinaryText, "", any);
        {error, Reason}  -> throw({error, file:format_error(Reason)})
    end.

%% @spec (binary(), string(), 'cr'|'lf'|'crlf'|'any') -> {string(), eol()}
%% @doc Canonicalizes the binary file text into a string.
%% <ol>
%%  <li>detects end-of-line marker style</li>
%%  <li>detects whether a final EOL was present</li>
%%  <li>appends a final `~n' EOL if not present</li>
%%  <li>replaces all EOLs found to `~n'</li>
%% </ol>
file_text(<<"\r\n", _/binary>>=Bin, Text, any) -> file_text(Bin, Text, crlf);
file_text(<<"\r",   _/binary>>=Bin, Text, any) -> file_text(Bin, Text, cr);
file_text(<<"\n",   _/binary>>=Bin, Text, any) -> file_text(Bin, Text, lf);

file_text(<<"\r\n">>, Text, crlf) -> {lists:reverse([$\n|Text]), {crlf, eol}};
file_text(<<"\r">>,   Text, cr)   -> {lists:reverse([$\n|Text]), {cr,   eol}};
file_text(<<"\n">>,   Text, lf)   -> {lists:reverse([$\n|Text]), {lf,   eol}};
file_text(<<>>,       Text, Eol)  -> {lists:reverse([$\n|Text]), {Eol,  noeol}};

file_text(<<"\r\n",Tail/binary>>, Text, crlf)-> file_text(Tail,[$\n|Text],crlf);
file_text(<<"\r",  Tail/binary>>, Text, cr)  -> file_text(Tail,[$\n|Text],cr);
file_text(<<"\n",  Tail/binary>>, Text, lf)  -> file_text(Tail,[$\n|Text],lf);
file_text(<<C,     Tail/binary>>, Text, Eol) -> file_text(Tail,[C  |Text],Eol).


%% @spec orig_text(eol(), Text) -> Text
%%       Text = [char() | Text]
%% @doc Restores the original whitespaces to the text stored inside graph
%% tokens.
orig_text(Eol, Text) when is_list(Text) ->
    [orig_text(Eol, El) || El <- Text];
orig_text({cr, _}, $\n)   -> $\r;
orig_text({crlf, _}, $\n) -> "\r\n";
orig_text({lf, _}, $\n)   -> $\n;
orig_text(_, Ch) when Ch >= 0, Ch =< 255 -> Ch;
orig_text(_, _) -> "".


%% @spec real_forms(node()) -> [node()]
%% @doc Returns the forms that are physically present in a file (i.e. skipping
%% the results of file inclusion).
real_forms(File) ->
    ?Query:exec(File, ?File:real_forms()).

%% @spec includers(node()) -> [node()]
%% @doc Returns files that depend on `File'.
includers(File) ->
    %% Probably these files should be sorted in a way that earlier files do
    %% not depend on later files. Currently this naive approach does not cause
    %% problems.
    ?Graph:path(File, [{incl,back}]) -- [File].


%% @doc Returns the output environment data stored in the graph representation.
%% There should be exactly one environment node; if it is not found,
%% `unsafe_save' is thrown.
output_env() ->
    case ?Graph:path(?Graph:root(), [{env, {name, '==', output}}]) of
        [Output] ->
            #env{value = Value} = ?Graph:data(Output),
            Value;
        _  ->
            throw(unsafe_save)
    end.

%%% ============================================================================
%%% Token level manipulation

%% @spec tokenize(string()) -> [formData()]
%% @doc Turns file text into a form list of token lists.
%% @todo lexical error handling
tokenize(Text) ->
    #fmenv{scanner = Scanner} = get(?EnvKey),
    case Scanner(Text, ?ErlScanner:init(fun make_token/4)) of
        {ok, Tokens} ->
            [{form_hash(Form), Form} ||
                Form <- split_forms(merge_ws(split_stop_tokens(Tokens)))];
        {error, {Ln, Mod, Error}, _Line} ->
            throw({Ln, Mod:format_error(Error)})
    end.

%% @spec store_tokens(formData()) -> inputForm()
%% @doc Store token data in the graph.
%% @todo probably ESG should be used to ensure garbage collection of nodes
store_tokens({Hash, Tokens}) ->
    {Hash,
     [{Token, ?Graph:create(#lex{type=token, data=Token})} || Token <- Tokens]}.


%% @doc Split `stop' tokens into a `stop' and a `ws' or `eol' token.
split_stop_tokens([#token{type=stop, text=Text} = Token | Tail]) ->
    [$.|WS] = Text,
    EOL = lists:last(WS),
    Type = if EOL == $\n; EOL == $\r -> eol; true -> ws end,
    [Token#token{text="."}, #token{type=Type, text=WS} |
     split_stop_tokens(Tail)];
split_stop_tokens([Token|Tail]) ->
    [Token|split_stop_tokens(Tail)];
split_stop_tokens([]) ->
    [].

%% @doc Put the contents of `ws' and `eol' tokens into the `prews' and `postws'
%% fields of the proper tokens.
merge_ws([#token{type=WST, text=WS, prews=Pre},
          #token{} = Token | Tail])
  when WST == ws; WST == eol ->
    merge_ws([Token#token{prews=Pre++WS} | Tail]);

merge_ws([#token{postws=Post} = Token | Tail]) ->
    case lists:splitwith(fun (#token{type=eol}) -> true;
                             (_) -> false
                         end, Tail) of
        {Eols, []} ->
            [Token#token{postws=Post ++ join_ws(Eols)}];
        {[], Rest} ->
            [Token | merge_ws(Rest)];
        {[Eol|Eols], Rest} ->
            {Head, Next} =
                lists:splitwith(
                  fun (#token{text=Txt}) ->
                          lists:any(fun(C)-> not lists:member(C," \t\n\r") end,
                                    Txt)
                  end, Eols),
            if
                Next =:= [] ->
                    [Token#token{postws = Post ++ Eol#token.text} |
                        merge_ws(Head ++ Next ++ Rest)];
                true ->
                    [Token#token{postws = Post ++ join_ws([Eol|Head])} |
                     merge_ws(Next ++ Rest)]
            end
    end.

%% @doc Join the textual contents of tokens.
join_ws(WS)                             -> join_ws(WS, []).
join_ws([#token{text=Text}|Tail], Join) -> join_ws(Tail, [Join, Text]);
join_ws([], Join)                       -> lists:flatten(Join).

%% @spec split_forms([#token{}]) -> [[#token{}]]
%% @doc Split a token list into forms.
split_forms([]) -> [];
split_forms(Tokens) ->
    {First, Rest} = split_first_form(Tokens),
    [First|split_forms(Rest)].

split_first_form(Tokens) ->
    split_first_form(Tokens, []).
split_first_form([Head=#token{type=stop}|Tail], Form) ->
    {lists:reverse([Head|Form]), Tail};
split_first_form([], Form) ->
    {lists:reverse(Form), []};
split_first_form([Head|Tail], Form) ->
    split_first_form(Tail, [Head|Form]).

%% =============================================================================

token_length(#token{prews=Pre, text=Text, postws=Post}) ->
    length(Pre) + length(Text) + length(Post).

set_form_hash(Form, Hash) when is_integer(Hash) ->
    case ?ESG:data(Form) of
        #form{hash=virtual} -> ok;
        #form{hash=Hash} -> ok;
        Data ->
            ?Graph:update(Form, Data#form{hash=Hash})
    end;
set_form_hash(Form, Text) when is_list(Text) ->
    set_form_hash(Form, form_hash(Text)).

%% @spec form_hash([#token{}] | Text) -> term()
%%       Text = [char() | Text]
%% @doc Returns the hash value corresponding to the given text that should be
%% stored for the form.
form_hash(Tokens = [#token{}|_]) ->
    form_hash(
      [[PreWS, Text, PostWS] ||
          #token{text=Text, prews=PreWS, postws=PostWS} <- Tokens]);
form_hash(Text) ->
    erlang:phash2(lists:flatten(Text)).

set_file_hash(File, Hash) when is_integer(Hash)->
    case ?ESG:data(File) of
        #file{hash=Hash}->ok;
        Data -> 
            ?ESG:update(File, Data#file{hash=Hash})
    end.

file_hash(File = {'$gn', file, _}) ->
    Forms = ?ESG:path(File, [form]),
    HashList = lists:filter(fun(X) -> X /= virtual end,
                            [(?ESG:data(Node))#form.hash || Node <- Forms]),
    erlang:phash2(HashList).

%% =============================================================================    

%% @spec merge(Old::[node()], New::[formData()]) -> [formAction()]
%% @doc Calculates a set of actions that update an `Old' form list to be the
%% same as the forms created from `New'. Tokens to be inserted are stored in
%% the graph during this process.
%% @todo Check if LCS algorithm is better suited. Sometimes it gives much
%% better results (e.g. when a form is moved from the end to the beginning),
%% but it is more costly (however, cutting the common prefix and postfix helps
%% in the usual case). It also needs to be considered wether preprocessor
%% changes can be taken into account with LCS (although a form-based
%% invalidation for preprocessor changes may make it irrelevant).
%%
%% Another note is that when a function is updated, its usually inserted
%% before the old one is deleted. In case of `graph_update', this approach
%% does not really work, analyser modules can't cope with that.

merge(Olds, News) ->
    case merge(Olds, News, []) of
        reload_all ->
            % todo It seems that if an include form is removed from an include file,
            %      the incl edges are not appropriately updated.
            [{del, Old} || Old <- Olds] ++
            [{ins, store_tokens(F)} || F <- News];
        Merge ->
            [Elem || Elem = {del, _} <- Merge] ++
            [Elem || Elem = {Tag, _} <- Merge, Tag =/= del]
    end.


merge(Old, [], R) ->
    lists:reverse(R, [{del, Form} || Form <- Old]);
merge([], New, R) ->
    lists:reverse(R, [{ins, store_tokens(Form)} || Form <- New]);
merge([OldForm | Old]=OL, [{NewHash, _} | New]=NL, R) ->
    #form{type=Type, tag=Tag, hash=OldHash} = ?Graph:data(OldForm),
    if
        Tag =:= include ->
            case is_include_form_changed(OldForm) of
                true ->
                    reload_all;
                false ->
                    merge(Old, New, [{hold, OldForm} | R])
            end;
        OldHash =:= NewHash ->
            merge(Old, New, [{hold, OldForm} | R]);
        Type =:= macro;
        Type =:= lex, (Tag =/= store) and (Tag =/= skip) ->
            reload_all_after(OL, R, NL);
        true ->
            merge_diff(OL, NL, R)
    end.

%% @doc When updating a file, a changed form is encountered.
%% Since the form may change how the later forms behave,
%% all forms after it (including the form itself) are completely reloaded.
reload_all_after(OL, R, NL) ->
    lists:reverse(R, [{del, F} || F <- OL] ++
                     [{ins, store_tokens(F)} || F <- NL]).


merge_diff(Old, [{Hash, _}=NewForm | New], R) ->
    case lists:splitwith(fun(F) -> form_hash_neq(F, Hash) end, Old) of
        {_AllDifferent, []} ->
            merge(Old, New, [{ins, store_tokens(NewForm)} | R]);
        {DiffPrefix, [OldForm | Rest]} ->
            merge(Rest, New,
                  [{hold, OldForm} |
                   lists:reverse([{del, F} || F <- DiffPrefix], R)])
    end.

%% @doc Returns whether the given form does NOT have the given hash.
form_hash_neq(Form, Hash) ->
    case ?Graph:data(Form) of
        #form{hash=Hash} -> false;
        _ -> true
    end.

%% @doc Returns whether the top-level include form
%% or any files included through it have been changed on disk
%% after its last change is recorded in the graph representation.
is_include_form_changed(IncForm) ->
    [TopIncFile] = ?Graph:path(IncForm, [iref]),
    IncFiles = ?Graph:path(TopIncFile, [incl]),
    lists:any(fun is_include_file_changed/1, IncFiles).

%% @doc Returns whether the file has been changed on disk
%% after its last change is recorded in the graph representation.
is_include_file_changed(IncFile) ->
    #file{lastmod=LastMod, path=Path} = ?Graph:data(IncFile),
    {ok, #file_info{mtime=ModifyTime}} = file:read_file_info(Path),
    ModifyTime > calendar:now_to_local_time(LastMod).


%% @spec hold_prefix(node(), integer()|last, [formAction()]) ->
%%                                            {[node()], [formAction()]}
%% @doc Constructs an action list that holds the first `Index' forms of
%% `File', and continues with `Actions'. Returns the list of real nodes as
%% well, because that will be used by `update/4'.
hold_prefix(File, Index, Actions) ->
    Prefix         = ?Graph:path(File, [{form, {1, Index}}]),
    RealForms      = real_forms(File),
    RealSet        = sets:from_list(RealForms),
    HoldRealPrefix = hold_real_prefix(Prefix, RealSet, Actions),
    {RealForms, HoldRealPrefix}.

hold_real_prefix([], _, Tail) -> Tail;
hold_real_prefix([Form|Rest], Real, Tail) ->
    case sets:is_element(Form, Real) of
        true  -> [{hold, Form} | hold_real_prefix(Rest, Real, Tail)];
        false -> hold_real_prefix(Rest, Real, Tail)
    end.


%% @spec input_tokens(node() | [node()]) -> inputForm()
%% @doc Returns token data either from an existing form or a list of existing
%% token nodes.

input_tokens(Tokens) when is_list(Tokens) ->
    TokenData = token_data(Tokens),
    {form_hash(Tokens), TokenData};

input_tokens(Form) ->
    #form{hash=Hash} = ?Graph:data(Form),
    Tokens = ?Syn:leaves(Form),
    RealHash =
    case Hash of
        undefined -> 
            form_hash(Tokens);
        _         ->
            Hash
    end,
    {RealHash, token_data(Tokens)}.

token_data(Tokens) ->
    [case ?Graph:data(Token) of
         #lex{type=token, data=D} -> {D, Token}
     end || Token <- Tokens].

dep_forms(Form) ->
    ?Graph:path(Form, [{fdep, back}]).

%% =============================================================================

action_length({ins, {_, [{    #token{type='-'}},
                         {    #token{type=atom,text="include"}},
                         {    #token{type='('}},
                         {Inc=#token{type=string}},
                         {    #token{type=')'}},
                         {    #token{type=stop}}]}}) ->
    IncludeFile = ?Token:get_value(Inc),
    incfile_length(IncludeFile);
action_length({ins, {_, Tokens}}) ->
    lists:sum([length(Text) || {#token{text=Text}, _} <- Tokens]);
action_length(_) ->
    0.

incfile_length(IncFile) ->
    {ok, Contents} = epp_dodger:parse_file(IncFile),
    SubLens = [incfile_length(SubInc)
                || {tree,attribute,
                    {attr,_,[],none},
                    {attribute,{atom,_,include},[{string,_,SubInc}]}} <- Contents
                    % not file_is_loaded(SubInc)
                    ],
    file_length(IncFile) + lists:sum(SubLens).

file_length(FileName) ->
    case filelib:is_file(FileName) of
        false ->
            0;
        true ->
            {ok, Binary} = file:read_file(FileName),
            binary_length(Binary)
    end.

binary_length(Binary) ->
    binary_length(Binary, 0).

binary_length(<<>>, N) -> N;
binary_length(<<_, Rest/binary>>, N) -> binary_length(Rest, N + 1).
%%binary_length(<<_, Rest/binary>>, N) -> binary_length(<<Rest>>, N + 1).

act_count(Actions)              -> act_count(Actions, 0).
act_count([{ins, _} | Tail], N) -> act_count(Tail, N+1);
act_count([{del, _} | Tail], N) -> act_count(Tail, N+1);
act_count([_        | Tail], N) -> act_count(Tail, N);
act_count([],                N) -> N.

%% =============================================================================

%% @spec file_status(node()) -> {file, node()} | {error, Reason}
file_status(File) ->
    case ?Graph:path(File, [{form, {type, '==', error}}]) of
        []    -> {file, File};
        [E|_] -> {error, (?Graph:data(E))#form.tag}
    end.



%% @doc Updates the last modification time of a file.
update_lastmod(File) ->
    FileData = ?Graph:data(File),
    ?ESG:update(File, FileData#file{lastmod=now()}).


%% @doc Creates a new backup file if the `file.erl' already exists
%% as described in the `backup' environment
%% (a list containing strings, 'datestamp' or 'timestamp'),
%% or `file.erl.bak.Date-Time' by default.
%% Backup is skipped if the `backup' environment contains `no_backup'.
%%
%% @todo Do something meaningful when the file copy fails.
make_backup(File) ->
    BackupStructure =
        case ?Syn:get_env(backup) of
            [no_backup] ->
                [];
            [Backup] ->
                Backup;
            _ ->
                ["bak", ".", datestamp, "-", timestamp]
        end,
    BackupPostfix = [convert_backup_str(BStr) || BStr <- BackupStructure],

    case {filelib:is_file(File), BackupPostfix} of
        {false, _} -> no_backup_needed;
        {true, []} -> no_backup_needed;
        {true, _}  ->
            NewFile = lists:flatten([File, "."] ++ BackupPostfix),
            file:copy(File, NewFile)
    end.

convert_backup_str(datestamp) ->
    {Y, Mo, D} = date(),
    io_lib:format("~4..0B~2..0B~2..0B", [Y, Mo, D]);
convert_backup_str(timestamp) ->
    {H, Mi, S} = time(),
    io_lib:format("~2..0B~2..0B~2..0B", [H, Mi, S]);
convert_backup_str(BStr) ->
    ?MISC:to_list(BStr).
    
%% =============================================================================

