%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2008-2009,
%%% Etvs Lornd University. All Rights Reserved.

%%% @doc File handler interface module for relative positioning. This module
%%% converts between the textual and graph representation of files 
%%% (parsing and saving).
%%%
%%% @author Laszlo Lovei <lovei@inf.elte.hu>
%%% @author Robert Kitlei <kitlei@inf.elte.hu>
%%% @author Gabor Hosszu

-module(refcore_fileman_rel).
-vsn("$Rev: 8864 $"). % for emacs"

%% Client interface / 
-export([add_file/1, add_file/2, drop_file/1, drop_file/2, save_file/1]).
-export([add_form/3, add_text/3, drop_form/2]).

%% gen_server callback functions
-export([start_link/0]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% used by ri:cat_errors/1
-export([create_scanner/0]).
-export([make_token/4,update/5,handle_add_form/3,handle_drop_form/2, 
         handle_add_text/3, handle_save_file/1,add_new_file/2, graph_update/2]).

-behaviour(gen_fileman).
-include("core.hrl").
-compile(export_all).
%%% ============================================================================
%%% Interface functions

%% @private
start_link() ->
	%?d(posmode_set),
	%?d(rel),
    gen_server:start_link({local, ?FILEMAN_SERVER}, ?MODULE, [], []).

%% @type progress() = (File::string(), Count::integer(), Max::integer()) -> ok.
%% A progress reporter function type that can be passed to {@link add_file/2}
%% and {@link drop_file/2}. This function is called to report the progress of
%% these operations. `File' is the name of the file on which progress has been
%% made, and there are `Count' steps finished a the time of the function call
%% from the total of `Max'.

-define(EnvKey, refcore_fileman_env).

%% This call style is used because we need these interface functions to work
%% when called recursively from the server process. In particular, the
%% preprocessor (which is called from the server process) must be able to add
%% include files to the graph.
-define(Call(Req), case get(?EnvKey) of
                       undefined ->
                           gen_server:call(?FILEMAN_SERVER, Req, infinity);
                       _ ->
                           handle(Req)
                   end).

%% @spec add_file(string()) -> {file, node()} | {error, Reason}
%%
%% @doc Reads and parses the file `Name' and inserts the syntax tree into
%% the graph. Returns the file node on success (which may contain forms with
%% errors). Note that during preprocessing, every included file will also be
%% added to the graph.
%% Note that this function does not save the graph to the disk.

add_file(Path) ->
    add_file(Path, []).

%% @spec add_file(string(), [Opt]) -> {file, node()} | {error, Reason}
%%
%% @doc The same as {@link add_file/1}, except that options can be passed to
%% modify the default behaviour. Currently supported options are:
%%
%% <dl>
%%
%%  <dt>{@type update | {update, true@}}</dt> <dd>Update the contents of
%%    `File' from disk, re-analysing only the forms that have been
%%    changed since the last update.</dd>
%%
%%  <dt>{@type {progress, progress()@}}</dt> <dd>Progress reporter to be
%%    called during analysis.</dd>
%%
%% </dl>
add_file(Path, Opts) when not is_list(Opts) ->
    add_file(Path, [Opts]);
add_file(Path, Opts) ->
    Update = proplists:get_value(update, Opts, false),
    Progress = proplists:get_value(progress, Opts, none),
    ?Call({add_file, Path, Update, Progress}).


%% @spec add_form(node(), integer() | last, [node()]) -> ok
%%
%% @doc Preprocesses `Tokens', creates a set of forms from the result, and
%% inserts these forms into `File' starting from the position `Index'.
%% `Index' specifies the position between links from `File' with tag `form'.
%% Should not be used to add a special form.
add_form(File, Index, Tokens) ->
    ?Call({add_form, File, Index, Tokens}).


%% @spec drop_form(node(), node()) -> ok
%% @doc Removes `Form' from `File', cleaning up preprocessor-related graph
%% parts.
drop_form(File, Form) ->
    ?Call({drop_form, File, Form}).


%% @spec add_text(node(), integer() | last, string()) -> ok
%%
%% @doc Turns `Text' into tokens, preprocesses them, creates a set of forms
%% from the result, and inserts these forms into `File' starting from the
%% position `Index'. `Index' specifies the position between links from `File'
%% with tag `form'.
add_text(File, Index, Text) ->
    ?Call({add_text, File, Index, Text}).


%% @spec drop_file(node()) -> ok
%%
%% @doc Removes `File' from the graph (together with all other files which
%% depend on it).
drop_file(File) ->
    drop_file(File, []).

%% @spec drop_file(string(), [Opt]) -> {file, node()} | {error, Reason}
%%
%% @doc The same as {@link drop_file/1}, except that options can be passed to
%% modify the default behaviour. Currently supported options are:
%%
%% <dl>
%%
%%  <dt>{@type {progress, progress()@}}</dt> <dd>Progress reporter to be
%%    called during analysis.</dd>
%%
%% </dl>
drop_file(FileName, Opts) ->
    Progress = proplists:get_value(progress, Opts, none),
    ?Call({drop_file, FileName, Progress}).


%% @spec save_file(node()) -> ok | {error, Reason :: string()}
%%
%% @doc Writes the textual contents of `File' back to its source.
save_file(File) ->
    ?Call({save_file, File}).


%%% ============================================================================
%%% Server callbacks

-record(fmenv, {scanner}).

%%% @private
init(_) ->
    create_scanner(),
    {ok, ?EnvKey}.

%%% @private
create_scanner() ->
    Scanner = ?ErlScanner:create(),
    put(?EnvKey, #fmenv{scanner=Scanner}).


%%% @private
handle_call(Req, _From, S) ->
    {reply, handle(Req), S}.

%%% @private
handle_cast(_, S) ->
    {noreply, S}.

%%% @private
handle_info(_, S) ->
    {noreply, S}.

%%% @private
terminate(_, _) ->
    ok.

%%% @private
code_change(_, S, _) ->
    {ok, S}.

%%% ============================================================================
%%% Implementation

%%% private type
%%% formData() = {Hash::integer(), [#token{}]}. Holds the tokens of a
%%% complete form before storing them in the graph. The hash value is used to
%%% compare with existing forms in the graph.

%%% private type
%%% inputForm() = {Hash::integer(), [{#token{}, node()}]}. Holds the
%%% tokens of a complete form after storing them in the graph. The hash value
%%% will be saved in the form node. The token list can be passed directly to
%%% the preprocessor, see {@link referl_preproc:formTokens()}.

handle({add_file, Info, Update, Progress}) ->
    Info2 =
        case Info of
            {New, Path} when is_list(Path) -> {New, ?MISC:canonical_filename(Path)};
            Path  when is_list(Path)       -> ?MISC:canonical_filename(Path);
            _                              -> Info
        end,
    handle_add_file(Info2, Update, Progress);
handle({add_form, File, Index, Tokens}) ->
    handle_add_form(File, Index, Tokens);
handle({drop_form, File, Form}) ->
    handle_drop_form(File, Form);
handle({add_text, File, Index, Text}) ->
    handle_add_text(File, Index, Text);
handle({drop_file, FileName, Progress}) when is_list(FileName) ->
    CFName = ?MISC:canonical_filename(FileName),
    handle_drop_file(CFName, Progress);
handle({drop_file, File, Progress}) ->
    handle_drop_file(File, Progress);
handle({save_file, File}) ->
    handle_save_file(File).

handle_add_file({New, File}, Update, Progress) ->
    case New of
        none ->
            case Update of
                true  -> disk_update(File, Progress);
                false -> {file, File};
                graph -> graph_update(File, Progress) % Internal use only
            end;
        new ->
            add_new_file(File, Progress)
    end;
handle_add_file(Path, Update, Progress) ->
    Info =
        case ?Graph:path(?Graph:root(), [{file, {path, '==', Path}}]) of
            [File] -> {none, File};
            []     -> {new, Path}
        end,
    handle_add_file(Info, Update, Progress).

handle_add_form(File, Index, Tokens) ->
    {Forms, Actions} = ?FileMan:hold_prefix(File, Index, 
                                            [{ins, ?FileMan:input_tokens(Tokens)}]),
    update(File, Forms, Actions, none, old). 

handle_drop_form(File, Form) ->
    Ind = ?Graph:index(File, form, Form),
    {RealForms, Actions} = ?FileMan:hold_prefix(File, Ind, [{del, Form}]),
    update(File, RealForms, Actions, none, old). 

handle_add_text(File, Index, Text) ->
    NewForms = [{ins, ?FileMan:store_tokens(T)} || T <- ?FileMan:tokenize(Text)],
    {RealForms, Actions} = ?FileMan:hold_prefix(File, Index, NewForms),
    update(File, RealForms, Actions, none, old). 


handle_save_file(File) ->
    FileData = #file{path=Path, eol=Eol} = ?ESG:data(File),
    SavePath =
        case ?FileMan:output_env() of
            original -> Path;
            Dir      -> filename:join(Dir, filename:basename(Path))
        end,
    ?FileMan:make_backup(SavePath),
    case file:open(SavePath, [write]) of
        {ok, Dev} ->
            RealForms  = ?FileMan:real_forms(File),
            FormWTexts = [{Form, ?Syn:flat_text(Form)} || Form <- RealForms],
            set_formlist_abs_pos(FormWTexts, 0),

            {_, FormTexts} = lists:unzip(FormWTexts),
            io:put_chars(Dev, ?FileMan:orig_text(Eol, FormTexts)),
            file:close(Dev),
            % This is used for file saving after undo
            ?ESG:update(File, FileData#file{lastmod=now()}),
            %update filehash 
            ?FileMan:set_file_hash(File, ?FileMan:file_hash(File)), 
            ?ESG:finalize(),
            ok;
        {error, Reason} ->
            {error, file:format_error(Reason)}
    end.

%%
%%	Make forms consistent: set its hash, and if needed recalculate tokens relative positions
%%    + set_form_hash
%%
%%	TODO: this function could carry a true/false atom, which could mean if there was a form 
%%		with improper hash
set_formlist_abs_pos([], _ActLen) ->
	ok;
set_formlist_abs_pos([ {Form, Text} | Tail ], ActLen) ->
	Hash = ?FileMan:form_hash(Text),
	case ?FileMan:form_hash_neq(Form,Hash) of
		false -> ok;
		true ->
			?FileMan:set_form_hash(Form, Hash),
			Toknodes = ?Syn:leaves(Form),
			Tokens = [ { extract_data(TNode) , TNode } || TNode <- Toknodes ],
			calc_rel_pos(Tokens,[])	
    end,
    set_form_abspos(Form, ActLen),
    set_formlist_abs_pos(Tail, ActLen + length(Text)).

%% @spec disk_update(node(), progress()) -> {file, node()}
%% @doc Updates the contents of `File' from its original disk file.
disk_update(File, Progress) ->
    FData = ?Graph:data(File),
    {Text, EOL} = ?FileMan:file_text(FData#file.path),
    ?Graph:update(File, FData#file{eol=EOL}),
    Forms = ?FileMan:real_forms(File),
    Merged = ?FileMan:merge(Forms, ?FileMan:tokenize(Text)),
    update(File, Forms, Merged, Progress, old),
    %% see remarks for graph_update
    [handle_add_file((?Graph:data(IF))#file.path, graph, Progress) ||
        IF <- ?FileMan:includers(File)],
    [?FileMan:update_lastmod(IF) || IF <- ?Graph:path(File, [incl])],
    ?ESG:finalize(),
    ?FileMan:file_status(File).

handle_drop_file(File, Progress) ->
    lists:foreach(
      fun (DepFile) ->
              Forms = ?FileMan:real_forms(DepFile),
              update(DepFile, Forms, [{del, F} || F <- Forms], Progress, old), 
              ?ESG:remove(?ESG:root(), file, DepFile)
      end, [File | ?FileMan:includers(File)]),
    ?ESG:finalize().

%% @spec add_new_file(string(), progress()) -> {file, node()}|{error, string()}
%% @doc Adds file from `Path' into the graph, assuming that it is not in the
%% graph yet.
add_new_file(Info, Progress) ->
    try
        case Info of
            Path when is_list(Path) ->
                {Text, EOL} = ?FileMan:file_text(Path),
                File = ?FileMan:create_file_node(Path, EOL);
            File ->
                {Text, EOL} = ?FileMan:file_text((?Graph:data(File))#file.path)
        end,
        ?Graph:update(File, (?Graph:data(File))#file{eol=EOL}),
        InsList = ?FileMan:tokenize(Text), 
        update(File, [],
               [{ins, ?FileMan:store_tokens(F)} || F <-InsList],
               Progress, new),
        ?ESG:finalize(),
        ?FileMan:file_status(File)
    catch
        {error, Error} ->
            {error, Error}
    end.


%% @spec graph_update(node(), progress()) -> {file, node()}
%% @doc Reparses the contents of `File' without rereading it from the disk.
%% @todo This should be replaced with a more fine-grained updating strategy
%% like invalidating only the referring forms when a preprocessor definition
%% is changed.
graph_update(File, Progress) ->
    Forms = ?FileMan:real_forms(File),
    Ins =
        [begin
             {Hash, OrigTokens} = ?FileMan:input_tokens(Form),
             {ins, ?FileMan:store_tokens({Hash, [T || {T, _} <- OrigTokens]})}
         end || Form <- Forms],
    update(File, Forms, [{del, F} || F <- Forms], Progress, old),
    ?ESG:finalize(),
    update(File, Forms, Ins, Progress, old),
    ?ESG:finalize(),
    ?FileMan:file_status(File).




%%% ============================================================================
%%% Form level manipulations

%%% private type
%%% formAction() = {ins, inputForm()} |
%%%                {hold, Form::node()} |
%%%                {del, Form::node()}

%% @spec update(node(), [node()], [formAction()], progress(), old|new) -> ok
%%
%% @doc Updates the contents of `File' by applying the elements of `Actions'.
%% `Forms' is the result of `real_forms(File)' (it is always computed before
%% calling `update'). `Progress' specifies what kind of progress reporting
%% should be done.

update(File, _Forms, Actions, Progress, OldOrNew) ->
    ALs = [?FileMan:action_length(Action) || Action <- Actions],
    {Lengths, TotalLength} = lists:mapfoldl(fun(X, S) -> {X + S, X + S} end, 0, ALs),
    #file{type=FileType, path=Path} = ?Graph:data(File),
    Now = os:timestamp(),
    Result = update(Actions, 1, start, File, FileType,
           ?FileMan:progress_start(Progress, Now, Path, Lengths, TotalLength, 
                                   ?FileMan:act_count(Actions)),
           OldOrNew, 0),
        %update filehash 
    ?FileMan:set_file_hash(File, ?FileMan:file_hash(File)),
    Result.

%%
%%	ActL is the absolute position of the current form
%% 
update([{hold, Form} | ATail], Index, PSt, File, FT, P, OldOrNew, ActL) ->
    Count = 1 + length(?FileMan:dep_forms(Form)),
    set_form_abspos(Form, ActL), %% TODO no update if not needed
    #form{length=Abs } = ?Graph:data(Form),
    [LTok] = ?Query:exec(Form, ?Syn:last_leaf()),
    #lex{ data=#token{ postws=PWS, sc=R } } = ?Graph:data(LTok),
    case R of
    	undefined ->
    		update(ATail, Index+Count, PSt, File, FT, P, OldOrNew, Abs );
    	_ -> { _ , EndP } = R,
    		 L = length(PWS) + EndP, 
    		update(ATail, Index+Count, PSt, File, FT, P, OldOrNew, Abs + L)
    end;
update([{del, Form} | ATail], Index, PSt, File, FT, P, OldOrNew, ActL) ->
    ?PreProc:detach(File, Form),
    [?ESG:remove(File, form, F) || F <- ?FileMan:dep_forms(Form)],
    ?ESG:remove(File, form, Form),
    update(ATail, Index, PSt, File, FT, ?FileMan:progress_step(P), OldOrNew, ActL);
update([{ins, {Hash, Input}} | ATail], Index, PSt, File, FT, P, OldOrNew, ActL) ->
	NInput = calc_rel_pos(Input,[]),
    {NewForms, PSt1} = ?PreProc:preprocess(NInput, File, PSt, OldOrNew),
    NewFormIdxs = ?MISC:index_list(NewForms, Index),
    %% run add on forms in NewFormIdxs
    NextL = add_forms_at(FT, File, Hash, Input, NewFormIdxs, ActL),
    update(ATail, Index + length(NewForms), PSt1, File, FT, ?FileMan:progress_step(P), OldOrNew, NextL);
update([], _Index, _PSt, _File, _FT, _P, _OldOrNew, _ActL) ->
    %% here we could shut down the progress reporter if needed
    ok.
%% Add forms from an Input
%% returns the start position of the next form (if it were after Input like this:
%%    Input|NextForm|NextForm2... )
add_forms_at(_FT, _File, _Hash, _Input, [], ActLength) ->
	ActLength;
add_forms_at(FT, File, Hash, Input, [ {ProcessedTokens, Index0} | Tail ], ActLength) ->
	NextLength = add_form_at(FT, File, Hash, Index0, ProcessedTokens, Input, ActLength),
	add_forms_at(FT, File, Hash, Input, Tail, NextLength).
%% Parses and adds the form at `Index' to the file.
%% returns the start position of the next form
add_form_at(FT, File, Hash, Index, ProcessedTokens, Input, ActLength) ->
    Form = ?FileMan:parse_form(File, Index, FT, ProcessedTokens), 
    ?FileMan:set_form_hash(Form, Hash),
    %% set abs pos (ActLength for the actual form, and calculate its length
    %% for the next form
    NextLength = calc_form_length(Form, ProcessedTokens, Input, ActLength),
    ?ESG:insert(File, {form, Index}, Form),
    ActLength + NextLength.

%% Calculate relative position for a list of tokens (starting at Acc)
calc_rel_pos(List,Acc) when is_list(List) ->
	calc_rel_pos(List,Acc,1,0);
calc_rel_pos(_List,_Acc) ->
 ok.

%% TODO N = ?
calc_rel_pos([],Acc,_Start,_N) ->
	lists:reverse(Acc);
calc_rel_pos([{Data, Node} | Tail ],Acc,Start,N) ->
	TStart = Start + length(Data#token.prews),
	TEnd = TStart + length(Data#token.text) - 1, 
	End = TEnd + length(Data#token.postws),
	H = Data#token{ sc={TStart,TEnd}  },
	?Graph:update(Node, {lex, token, H  }),
	calc_rel_pos(Tail, [ { Data#token{ sc={Start,End} } , Node } | Acc ] ,End+1,N+1).

extract_data(TNode) ->
	{ _, _, D} = ?Graph:data(TNode),
	D.


%% Set absolute position for a form
set_form_abspos(Form, Abs) -> 
    ?Graph:update(Form, (?Graph:data(Form))#form{length=Abs }).
%% Calculate form length (and set its absolute position)
%% This function returns the calculated form length
calc_form_length(Form, {vtokens, Dep, _Orig, _Tokens}, _Input, _ActLength) ->
  	#form{ length=Abs } = ?Graph:data(Dep),
  	set_form_abspos(Form, Abs),
    0;
calc_form_length(Form, _, TokenWPoss, ActLength) ->
    NextLength = lists:sum([?FileMan:token_length(Token) || {Token, _} <- TokenWPoss]), 
    set_form_abspos(Form, ActLength),
    NextLength.


%%% ============================================================================
%%% Token level manipulation

%% Create a token node
make_token(Type, Text, {_LStart, _LEnd}, {_SSt, _SEnd}) ->
    %% TODO: no ?Token call (move ?FileMan:build into ?Syn)
    ?Token:build(Type, Text).


%% ===========================================================================================
%% Makes the error text more informative in cases when it is possible.
improve_msg(File, _Line, FormIndex, _Tokens, ["syntax error before: ", ErrPosMsg]) ->
    FlatTxt = ?MISC:format(ErrPosMsg) ++ ".",
    case ?MISC:string_to_term(FlatTxt) of
    	{ok, {_ErrToken = #token{text=Text, type=Type, sc={B,E}}, _ErrLex}} ->
			case Type of
				'?' ->
				    % a macro for which no substitution could be found
				    % as it is already reported, no further info message is required
				    "";
				_ ->
				    #file{path=Path} = ?Graph:data(File), ?d(FormIndex),
				    LastFormIndex = FormIndex - 1,
				    Length =
				    case LastFormIndex > 0 of
				    	true -> [LastForm] = 
				    				?Query:exec(File, ?File:form(LastFormIndex)),
				    	        LastFormData =?Graph:data(LastForm),
				    			#form{ length=LastLength } = LastFormData,
				    			GetLastLeaf = ?Syn:last_leaf(),
								[LastLeaf]    = GetLastLeaf(LastForm),
								#lex{ data=#token{ sc={_,Pos}, postws=PostWs } }
														  = ?ESG:data(LastLeaf),
								LastLength + Pos + length(PostWs);
				    	_	 ->	0
				    end,
				    {BS,ES} = { Length + B , Length + E },
				    FileData = file:read_file(Path),
				    [Beg, _End] = ?Token:map_pos2lc(FileData, [BS,ES]),
				    {ErrLine, ErrCol} = Beg,

				    Text2 = ?MISC:format(Text),
				    Text3 = 
				     re:replace(Text2, "\n", "\\\\n", [global, {return, list}]),
				    ?MISC:format("~s:~p:~p Parse error before " ++ Text3 ++ "~n",
				                         [Path, ErrLine, ErrCol])

			end;
		{error,{_Num,_ErrAtom,ErrText}} ->
    		?MISC:format(lists:flatten(ErrText) ++ "\n");
    	_Other ->
     		?MISC:format("Syntax error.\n")	
	end;
		
improve_msg(_File, _Index, Mod, _Tokens, Msg) ->
    ?MISC:format(Mod:format_error(Msg) ++ "\n").


token_to_linecol(#token{prews=Pre, text=Text, postws=Post}) ->
    RevFullText = lists:reverse(Pre ++ Text ++ Post),
    {LastLine, LinesStr} = lists:splitwith(fun(Ch) -> Ch =/= $\n end, RevFullText),
    LineCnt = lists:sum([1 || $\n <- LinesStr]),
    {LineCnt, length(LastLine)}.

add_linecols([], LineCol) ->
    LineCol;
add_linecols([undefined|LineCols], LineCol) ->
    add_linecols(LineCols, LineCol);
add_linecols([{0, CurrCol} | LineCols], {Line, Col}) ->
    add_linecols(LineCols, {Line, Col + CurrCol});
add_linecols([{CurrLine, CurrCol} | LineCols], {Line, _Col}) ->
    add_linecols(LineCols, {Line + CurrLine, CurrCol}).
