%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2008-2009,
%%% Etvs Lornd University. All Rights Reserved.

%%% @doc UI message router
%%% @author bkil.hu
%%% @author Viktoria Fordos <f-viktoria@elte.hu>

%%% The strategy for managing requests
%%% When a new request is arrived we examine it, whether is it a new request.
%%% A request, which is not a member of the requests list, is called new request.
%%% Otherwise, it is called  registered request. This classification
%%% is needed, because of some ui services (etc refactorings) exist, 
%%% which require more then one steps for its execution 
%%% and require user interactions.
%%%
%%% If a request is a registered request, its execution is always permitted
%%% and any administration is not needed, because of its parent request 
%%% had been administrated.
%%%
%%% If a request is a new request, another examinations would be performed.
%%% First, we have to determine whether the request is not in the 'ui router's
%%% blacklist'. If it is,it should be denied.
%%% The 'ui router's blacklist' is a list of denied operations. These operations
%%% can not be allowed to run in any circumstances. The blacklist is loaded from
%%% a file during the initialization of the server.
%%%
%%% Second, we have to determine the type of the request first. It can be a modifier or 
%%% a non-modifier. If a request modify the database, then it is called modifier, 
%%% else it is called non-modifier. After the type of the request has been 
%%% determined, the permission of the execution should be limited, as described 
%%% below.
%%% A modifier can only be executed when nothing else is running. 
%%% A non-modifier can only not be executed when a modifier is running. 
%%%
%%% If the request owns the permit execution then the following sequence 
%%% should be executed.
%%% 1) Before the execution of the request, we add it to the requests list.
%%% 2) The request is executed.
%%% 3) After the execution of the request, we send back its result for 
%%% the client, and remove it from the requests list.
%%%
%%% If the request has not got the permit execution then it should be denied 
%%% and should be ignored.

-module(reflib_ui_router).
-vsn("$Rev: 8597 $ ").
-behaviour(gen_server).

%-compile(export_all).

%% interface
-export([getid/0, request/2, request/3]).
-export([add_msg_handler/1, del_msg_handler/1]).

%% private
-export([broadcast/2]).

%%% Server exports
-export([start_link/0]).

%% gen_server callbacks
-export([handle_call/3, handle_cast/2]).
-export([init/1, handle_info/2, terminate/2, code_change/3]).

%% Callbacks
-export([error_text/2]).

-export([set_ref_node/1]).

%-define(DEBUG, true).
 
-ifdef(DEBUG).
-define(debug(Var), 
        error_logger:info_msg("DEBUG: ~p:~p~n~p~n  ~p~n", 
                              [?MODULE, ?LINE, ??Var, Var])).
-else.
-define(debug(__Arg), ok).
-endif.

-include("lib.hrl").

%% constants
-define(T_Broadcast, global).
-define(T_Reply, reply).

-define(NoReply,   noreply).
-define(ERR(R),    {error,{R, ?Error:error_text(R)}}).
-define(LErr(R),   ?LocalErr(R,[])).
-define(LocalErr(R,L), ?ERR(?LocalError(R,L))).
-define(RefErr(R,L),   ?ERR(?RefError(R,L))).

%Filename of the blacklist
-define(BLACKLIST_FILE, filename:join(mnesia:system_info(directory),
                                      "ui_router_blacklist")).

%% State of the ui_router
-record(state, {
                % Message handlers
                msg_handlers = []     :: list(),
                % Requests 
                requests = []  :: list(),
                % Blacklist
                blacklist = [] :: list()
               }).

%% Type of the available UI requests.
%% If a request modify the database, then it is called modifier, 
%% else it is called non-modifier.
%% Every UI request, which is identified by its name, needs to be classify.
%% The classes are exclusive to each other.
-record(request_type, {
                     %List of modifiers
                     modifiers=[%Database operations 
                                stop, reset, clean, add, drop, backup,
                                %Graph operations 
                                %not implemented in refdb_mnesia, but implemented in refdb_nif
                                create_graph, rename_graph, restore, undo, load_graph,
                                delete_graph, delete_all_graphs,
                                %Handling beam files
                                load_beam,
                                %Database operations (user level)
                                add_dir, generate_dir, drop_dir,
                                %Enviromental nodes
                                add_env, del_env, del_env_val, set_env,
                                %Clustering
                                run_cl, cl_refresh,
                                %Database synchronization
                                synchronize],


                     %List of non-modifiers
                     non_modifiers=[%Database operations
                                    ls_backups, backup_info,
                                   %Graph operations
                                    ls_graphs, actual_graph,
                                    %Status queries
                                    status, showconfig, filelist, status_info,
                                    cat_file, cat_fun, cat_recmac,
                                    %Enviromental nodes
                                    get_envs, get_env,
                                    %Dep. anal.s
                                    cyclic_fun_check_cycle, cyclic_fun_check_cycle,
                                    cyclic_fun_print_cycle, cyclic_fun_print_cycle,
                                    cyclic_fun_check_function, cyclic_fun_draw,
                                    cyclic_fun_draw,cyclic_fun_draw_cycles, 
                                    cyclic_fun_draw_cycles,cyclic_mod_check_module,
                                    cyclic_mod_print_cycle, cyclic_mod_print_cycle,
                                    cyclic_mod_check_cycle, cyclic_mod_check_cycle,
                                    cyclic_mod_draw, cyclic_mod_draw, 
                                    cyclic_mod_draw_cycles, cyclic_mod_draw_cycles,
                                    %Directory sorting
                                    dir_sort, dir_sort_get_dirs, dir_sort_get_mods,
                                    %Function block analyses
                                    refusr_fb_regexp_re,fb_relations_get_relations,
                                    fb_relations_is_relation, fb_relations_draw,
                                    fb_relations_check_cycle, fb_relations_draw_cycles,
                                    %Interface layers
                                    if_layers_show_insults, if_layers_draw, 
                                    if_layers_draw, 
                                    if_layers_check_layered_architecture,
                                    %Duplicated code
                                    duplicated_code_initial_clone, 
                                    duplicated_code_const_var_diff,
                                    %Semantic queries utils
                                    draw, get_running_queries, kill_query,
                                    funlist, recordlist, macrolist,
                                    %Transformation utils
                                    reply, cancel,
                                    %Clustering
                                    cl_options,
                                    %Metric mode
                                    metric_mode,
                                    transform,
                                    %graph querying
                                    graph_query, graph_data,
                                    syn_leaves, syn_index,
                                    %html generation
                                    html_generate_node, html_generate,
                                    database_hash, file_hash,
                                    current_posmode]
                    }).


%% Request data
-record(request, {%Identify the request
                  reqid :: string(),
                  %The type of the request
                  req_type :: modifier | non_modifier
                  }).



%%% ============================================================================
%%% Error texts

error_text(bad_request, [X]) ->
    ["Bad UI request: ", ?MISC:any_to_string(X)];
error_text(exception, [Cl,Error,Stack]) ->
    ["Unhandled exception ", atom_to_list(Cl), ":", io_lib:write(Error),
    io_lib:print(Stack)];
error_text(bad_response, [X]) ->
    ["Bad UI functional backend response: ", ?MISC:any_to_string(X)].

%%% ============================================================================
%%% Error texts

% Note: you MUST NOT rely on the internal structure of a request ID
getid() ->
    gen_server:call({?MODULE, ?REFERL_NODE}, {getid}).

% Note: you MUST NOT rely on the internal structure of a request ID
request(ReqId, Request) ->
    request(self(), ReqId, Request).

% Note: you MUST NOT rely on the internal structure of a request ID
request(Pid, ReqId, Request) when is_pid(Pid) ->
    RPid = {Pid, node()},
    gen_server:call({?MODULE, ?REFERL_NODE}, {request, RPid, ReqId, Request});

% Note: you MUST NOT rely on the internal structure of a request ID
request(RPid, ReqId, Request) ->
    gen_server:call({?MODULE, ?REFERL_NODE}, {request, RPid, ReqId, Request}).

add_msg_handler(Pid) ->
    gen_server:call({?MODULE, ?REFERL_NODE}, {add_msg_handler, Pid}).

del_msg_handler(Pid) ->
    gen_server:call({?MODULE, ?REFERL_NODE}, {del_msg_handler, Pid}).

broadcast(Type, Details) ->
    gen_server:cast({?MODULE, ?REFERL_NODE}, {broadcast, Type, Details}).

%%

%% @doc Starts the server and links it with the calling process.
start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).


init(_Args) ->
    process_flag(trap_exit, true),
    Blacklist = init_blacklist(),
    case check_exclusive_constraint() of
        ok -> {ok, #state{blacklist = Blacklist}};
        violator_found -> {stop, internal_error}
    end.

set_ref_node([server, _ ])->
    set_ref_node([node()]);

set_ref_node([Node]) ->
    application:set_env(refactorerl, referl_node, Node),
    ok.

%% ================
%% Implementation

handle_call({getid}, _From, State) ->
    Ref = make_ref(),
    Reply = {reqid, base64:encode_to_string(term_to_binary(Ref))}, %@todo
    {reply, Reply, State};


handle_call({request, {Pid, Node}, ReqId={reqid,S}, Request}, _From, State)
  when is_pid(Pid), is_atom(Node), is_list(S) ->
    {Reply, NewState}=case is_new_request(S, State#state.requests) of
        true -> start_new_request(Pid, S, Request, State);
        false -> R = handle_registered_request(Pid, ReqId, Request),
                 {R, State}
    end,
    ?debug(NewState),
    {reply, Reply, NewState};

handle_call({request, _Pid, _ReqId, _Request}, _From, State) ->
    {reply, error, State};

handle_call({add_msg_handler, HandlerPid}, _From, 
            State = #state{msg_handlers=Handlers} ) ->
    HandlerPid ! installed,
    NewHandlers = [HandlerPid | Handlers],
    {reply, ok, State#state{msg_handlers= NewHandlers}};

handle_call({del_msg_handler, Handler, _Args}, _From, 
            State = #state{msg_handlers=Handlers} ) ->
    NewHandlers = lists:delete(Handler, Handlers),
    {reply, ok, State#state{msg_handlers= NewHandlers}};

handle_call({remove_request, ReqId}, _From, State = #state{requests = Requests})
  when is_list(ReqId) ->
    NewRequests = lists:filter(fun(#request{reqid = Id}) -> 
                                       (Id /= ReqId) 
                               end,
                               Requests ),
    ?debug(NewRequests),
    {noreply, State#state{requests=NewRequests}};


handle_call(_, _, State) ->
    {noreply, State}.

handle_cast({broadcast, Type, Details}, State = #state{msg_handlers=Handlers} ) ->
    lists:foreach(fun(Pid) ->
                          Pid ! {?T_Broadcast, Type, Details}
                  end, Handlers),
    {noreply, State};

handle_cast(_, State) ->
    {noreply, State}.

handle_info({'EXIT', Pid, _Reason}, State = #state{msg_handlers=Handlers} ) ->
    NewHandlers = lists:delete(Pid, Handlers),
    {noreply, State#state{msg_handlers= NewHandlers}};

%% @private
handle_info(_, S) ->
    {noreply, S}.

%% @private
terminate(_,_) ->
    ok.

%% @private
code_change(_,S,_) ->
    {ok, S}.

%% ========
%% Private implementation

create_callbacks(Pid, ReqId) ->
    UnicastFun =
        fun(Type, Details) -> 
            Pid ! {ReqId, Type, Details}
        end,
    BroadcastFun =
        fun(Type, Details) ->
            ?MODULE:broadcast(Type, Details)
        end,
    #msg_cb{unicast = UnicastFun, broadcast = BroadcastFun}.

run(Request, MCB=#msg_cb{},RunAfter)->
    spawn(fun() ->
                  R = exec_req(Request, MCB),
                  case R of
                      ?NoReply -> ?NoReply;
                      _ ->
                          Reply =
                              case R of
                                  {ok,{abort,E}} ->
                                      {ok,{abort,{E,?Error:error_text(E)}}};
                                  {ok,_}->
                                      R;
                                  {error,E} ->
                                      ?ERR(E);
                                  _ ->
                                      ?LocalErr(bad_response, [R])
                              end,
                          (MCB#msg_cb.unicast)(?T_Reply, Reply),
                          RunAfter()
                  end
          end),
    ok.

exec_req(Request, MCB=#msg_cb{}) ->
    M = ?UIB,
    UI_funs = proplists:get_value(exports, M:module_info()),
    case catch tuple_to_list(Request) of
        [F | Args0] ->
            Args = [MCB | Args0],
            FunArity = {F,length(Args)},
            case lists:member(FunArity, UI_funs) of
                true ->
                    try
                        apply(M, F, Args)
                    catch
                        error:Error = {_,_,_} ->
                            {error, Error};
                        throw:Error = {_,_,_} ->
                            {error, Error};
                        Cl:Error ->
                            {error,?LocalError(exception, [Cl, Error,
                                                  erlang:get_stacktrace()])}
                    end;
                false ->
                    {error,?LocalError(bad_request,[Request])}
            end;
        _ ->
            {error,?LocalErr(bad_request,[Request])}
            %@todo: or perhaps throw an exception instead?!
    end.

add_request(Request, State = #state{requests = Requests}) 
  when is_record(Request, request) ->
    NewRequests = Requests ++ [Request],
    State#state{requests=NewRequests}.
    
remove_request(ReqId)->
    gen_server:call({?MODULE, ?REFERL_NODE}, {remove_request, ReqId}).

start_new_request(Pid, ReqId, Request, State)->
    #state{requests = Requests, blacklist = Blacklist} = State,
    RequestType = determine_request_type(Request),
    case not_in_blacklist(Request, Blacklist) 
        andalso can_permit_execution(RequestType, Requests) of
        false -> {deny, State};
        true -> Result = run_request(Pid, ReqId, Request, true),
                NewState = add_request(#request{reqid = ReqId,
                                                req_type = RequestType},
                                       State),
                {Result, NewState}
    end.

handle_registered_request(Pid, ReqId, Request)->
    run_request(Pid, ReqId, Request, false),
    ok.

run_request(Pid, ReqId, Request, NeedRemove)->
    MCB = create_callbacks(Pid, {reqid,ReqId}),
    Fun = fun()->
                  case NeedRemove of
                      true -> remove_request(ReqId);
                      _ -> ok
                  end
          end,
    run(Request, MCB, Fun).

% It verifies whether the available UI requests are complying with the rule.
% The rule is descirbed above.
% Every UI request is assigned to exactly one class {modifier, non_modifier}.
check_exclusive_constraint()->
    RequestType=#request_type{},
    Modifiers = RequestType#request_type.modifiers,
    NonModifiers = RequestType#request_type.non_modifiers,
    case sets:is_disjoint(sets:from_list(Modifiers), 
                          sets:from_list(NonModifiers)) of
        true -> ok;
        false -> violator_found
    end.

% If the request Id of the request can not be be found in the requests list, 
% then the request is called new request.
is_new_request(ReqId, Requests) when is_list(ReqId) andalso is_list(Requests)->
    case lists:keyfind(ReqId, #request.reqid, Requests) of
        R when is_record(R, request) -> false;
        _ -> true
    end.

% We assume that a request is a modifier 
% till we can not proof that the inverse property is true. 
% If the request is an transformation we handle it as an exception.
determine_request_type({transform, Action, _}) ->
	case ?UIB:is_transformation(?UIB:action_module_name(Action)) of
		true -> modifier;
		false -> non_modifier
	end;
determine_request_type(Request) when is_tuple(Request) ->
    determine_request_type(element(1, Request));
determine_request_type(Request) when is_atom(Request) ->
    case lists:member(Request, (#request_type{})#request_type.non_modifiers) of
        true -> non_modifier;
        false -> modifier
    end.

% A modifier can only be executed when nothing else is running.
% A non-modifier can only not be executed when a modifier is running.
can_permit_execution(modifier, [])->
    true;
can_permit_execution(modifier, _) ->
    false;
can_permit_execution(non_modifier, [ #request{req_type = modifier} | _Tail ]) ->
    false;
can_permit_execution(non_modifier, _) ->
    true.

% If the blacklist consists the request,  then we say, the request is in blacklist.
not_in_blacklist(Request, Blacklist) when is_tuple(Request) ->
    not_in_blacklist(element(1, Request), Blacklist);
not_in_blacklist(Request, Blacklist) when is_list(Blacklist) ->
    not lists:member(Request, Blacklist).
   
% The initialization of the blacklist.
init_blacklist()->
    case file:consult(?BLACKLIST_FILE) of
        {ok, [BlackList]} when is_list(BlackList) ->
            lists:filter(fun(E)->
                                 is_atom(E) 
                         end, BlackList);
        _ -> []
    end.
