%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2009,
%%% Etvs Lornd University. All Rights Reserved.

%%% @author Viktoria Fordos <v@fviktoria.hu>

-module(referl_ui_nitrogen_skeletons).

-export([update_skeleton/3, update_prev_skeleton_comment/2, delete_skeleton/1,
         skeleton_call_format/1, save_skeleton/3, list_skeletons/0, 
         evaluate_skeleton/3, evaluate_skeleton/4, try_parse_skeleton/2,
         determine_sq_request_type/1, do_autocomplete_skeleton/1]).

-define(SPARAM_START, $$).
-define(SPARAM_END, $$).

-define(NITRO_ERROR, nitrogen_helper_error).
-vsn("$Rev: 8750 $ ").

-include("ui.hrl").
-include_lib("referl_core/include/core_export.hrl").

%%% ============================================================================
%%% Skeletons
%% @doc Replaces the body and the owner of the skeleton, 
%% which is identified by the given name,
%% with the given new body and with the given owner.
%% @spec update_skeleton(Name::string(), NewBody::string(), Owner::string())->
%% ok | {error, ErrorMessage::string()}
update_skeleton(Name=[A|_],NewBody=[B|_],Owner=[C|_]) when is_integer(A) andalso
    is_integer(B) andalso is_integer(C) ->
    try
        check_skeleton_syntax(NewBody),
        [{_,_,_,_,Comment}] = ?NITRO_HELPER:find_in_stab_by_name(Name),
        save_safe_skeleton(Name, NewBody, Owner, Comment)
    catch
        {?NITRO_ERROR, ErrorMessage}->{error, ErrorMessage};
        _:_ -> {error, "Unknown error occured"}
    end.

%% @doc Updates the comment of a previously stored skeleton based on 
%% the given skeleton name.
%% @spec update_prev_skeleton_comment(Name::string(), NewComment::string())-> ok
update_prev_skeleton_comment(Name=[A|_], NewComment) when is_integer(A)
    andalso is_list(NewComment)->
    [{_,Body,Owner,_Card,_}] = ?NITRO_HELPER:find_in_stab_by_name(Name),
    save_safe_skeleton(Name, Body, Owner, NewComment).

%% @doc Deletes the skeleton which is identified by the given name
%% @spec delete_skeleton(Name::string())-> ok
delete_skeleton(Name)->
    ?NITRO_HELPER:delete_from_stab({Name, '_', '_', '_', '_'}).

%% @doc Composes a valid skeleton call based on the given skeleton name.
%% @spec skeleton_call_format(Name::string())->string()
skeleton_call_format(Name=[C|_]) when is_integer(C)->
    [{_,_,_,Card,_}] = ?NITRO_HELPER:find_in_stab_by_name(Name),
    ParamStr=case Card of
                 0 -> "";
                 1 -> " `_` ";
                 N -> string:copies(" `_` ,", N-1)++" `_` "
             end,
    Name++"("++ParamStr++").".

%% @doc Saves a new skeleton from the given parameters.
%% @spec save_skeleton(Name::string(), Body::string(), Owner::string())->
%% ok | {error, ErrorMessage::string()}
save_skeleton(Name=[A|_], Body=[B|_], Owner=[C|_]) when is_integer(A) andalso
    is_integer(B) andalso is_integer(C) ->
    case validate_skeleton(Name, Body) of
        ok -> save_safe_skeleton(Name, Body, Owner, "");
        Err={error, _} -> Err
    end.

%% @doc Returns a list of the data of the stored skeletons.
%% @spec list_skeletons()->[]|[Skeleton]
%% Skeleton = {Name, Body, Owner, ParamCardinality, Comment}
%% Name = string()
%% Body = string()
%% Owner = string()
%% ParamCardinality = integer()
%% Comment = string()
list_skeletons()->
    ?NITRO_HELPER:get_stab_elements().

%% @doc Evaluate the skeleton, which is identified by the given name, with the
%% given actual parameters than try to execute it as a semantic query string.
%% @spec evaluate_skeleton(Name::string(), Parameters, User::string()) ->
%% {Result, Warning, Error}
%%   Result = {result, no_result | [term()]}
%%   Warning = {warning, no_warning | string()}
%%   Error = {error, no_error | string()}
%%   Parameters = []| [string()]
evaluate_skeleton(Name=[A|_], Parameters, User=[B|_]) when is_integer(A) 
  andalso is_integer(B) andalso is_list(Parameters)->
    try
        case ?NITRO_HELPER:find_in_stab_by_name(Name) of
            [] ->throw({?NITRO_ERROR, "Skeleton was not found."});
            [Result={_, Body, _, _, _}] -> 
                check_skeleton_params_match(Result, Parameters),
                Query=convert_to_sq(Body, Parameters),
                ?NITRO_SERVICES:execute_query({'query', Query, User})
        end
    catch
        {?NITRO_ERROR, ErrorMessage}->{{result, no_result},
                                       {warning,no_warning},
                                       {error,ErrorMessage}};
        {_, _} -> {{result, no_result},
                   {warning,no_warning},
                   {error,"Unknown error occured"}}
    end.

evaluate_skeleton(Name=[A|_], Parameters, _User, onlyconvert) when is_integer(A) 
  andalso is_list(Parameters)->
    try
        case ?NITRO_HELPER:find_in_stab_by_name(Name) of
            [] ->throw({?NITRO_ERROR, "Skeleton was not found."});
            [Result={_, Body, _, _, _}] -> 
                check_skeleton_params_match(Result, Parameters),
                convert_to_sq(Body, Parameters)
        end
    catch
        {_, _} -> error
    end.

%%% ============================================================================
%%% helper functions
do_autocomplete_skeleton(Call)->
    lists:foldl(fun({Name,_,_,_,_}, Acc)->
                    FSkel = skeleton_call_format(Name),
                    case lists:prefix(Call, FSkel) of
                        true-> Acc++[{FSkel, FSkel}];
                        false -> Acc
                    end
                end, [], lists:keysort(1, list_skeletons())).
    
determine_sq_request_type(Call=[C|_]) when is_integer(C)->
    try
        check_skeleton_call_syntax(Call),
        skeleton_name_registered(Call),
        skeleton
    catch
        {_, _} -> sem_query
    end.

try_parse_skeleton({Call, _File, _Pos}, User)->
    {ok, Elements, _}=erl_scan:string(Call),
    {_,_, SkelName}=hd(Elements),
    ParamsList=re:split(Call,"[``]",[{return,list}]),
    Parameters=case length(ParamsList) of
                   1 ->[]; %skeleton can contain 0 parameters
                   _ -> ?NITRO_HELPER:even_list_elements(ParamsList)
               end,
    evaluate_skeleton(atom_to_list(SkelName), Parameters, User).

check_skeleton_params_match({_,_,_,ParamCard,_}, Params) when is_list(Params)->
    case length(Params)==ParamCard of
        true-> ok;
        false -> throw({?NITRO_ERROR, 
                        "Cardinality of the actual parameters does not 
                        match the cardinality of the formal parameters."})
    end.

convert_to_sq(Body, []) -> Body;

convert_to_sq(Body=[A|_], [HParam|Params]) when is_integer(A) 
  andalso is_list(HParam) andalso is_list(Params)->
    Pattern="(["++[?SPARAM_START]++"][^"++[?SPARAM_END]++
            "]*["++[?SPARAM_END]++"])",
    {ok, Mp}=re:compile(Pattern,[]),
    NewBody=re:replace(Body,Mp,HParam,[{return, list}]),
    convert_to_sq(NewBody, Params).

save_safe_skeleton(Name=[A|_], Body=[B|_], Owner=[C|_], Comment) 
    when is_integer(A) andalso is_integer(B) 
    andalso is_integer(C) andalso is_list(Comment)->
    ParamCardinality=countSkeletonParameters(Body),
    ?NITRO_HELPER:insert_to_stab(Name, Body, Owner, ParamCardinality, Comment).

validate_skeleton(Name=[A|_], Body=[B|_]) 
    when is_integer(A) andalso is_integer(B)->
    try
        check_skeleton_syntax(Body),
        has_unique_skeleton_name(Name),
        ok
    catch
        {?NITRO_ERROR, ErrorMessage}->{error, ErrorMessage};
        {_, _} -> {error, "Unknown error occured"}
    end;

validate_skeleton(_,_)->
    {error, "Bad args"}.

%This clause should be used when the start and end symbols 
%of the skeleton's parameters are not equal to each other.
%countSkeletonParameters(Body=[A|_]) 
%     when is_integer(A) andalso ?SPARAM_START/=?SPARAM_END ->
%    StartSymbols=lists:filter(fun(E)->E==?SPARAM_START end, Body),
%    length(StartSymbols);

countSkeletonParameters(Body=[A|_]) 
    when is_integer(A) andalso ?SPARAM_START==?SPARAM_END ->
    StartSymbols=lists:filter(fun(E)->E==?SPARAM_START end, Body),
    length(StartSymbols) div 2.

%This clause should be used when the start and end symbols of the 
%skeleton's parameters are not equal to each other.
%check_skeleton_syntax(Body=[A|_]) 
%     when is_integer(A) andalso ?SPARAM_START/=?SPARAM_END->
%    StartSymbols=lists:filter(fun(E)->E==?SPARAM_START end, Body),
%    EndSymbols=lists:filter(fun(E)->E==?SPARAM_END end, Body),
%    case length(StartSymbols)==length(EndSymbols) of
%        true-> ok;
%        false -> throw({?NITRO_ERROR,
%                        "Cardinality of the start symbol does not match
%                          the cardinality of the end symbol."})
%    end;

check_skeleton_syntax(Body=[A|_]) 
    when is_integer(A) andalso ?SPARAM_START==?SPARAM_END->
    Symbols=lists:filter(fun(E)->E==?SPARAM_START end, Body),
    case ?NITRO_HELPER:even(length(Symbols)) of
        true-> ok;
        false -> throw({?NITRO_ERROR,
                        "Cardinality of the start symbol does not match
                         the cardinality of the end symbol."})
    end.

skeleton_name_registered(Call)->
    Name=lists:takewhile(fun( $( )->false; (_)-> true end, Call),
    case length(?NITRO_HELPER:find_in_stab_by_name(Name)) of
        0 -> throw({?NITRO_ERROR, Name++" does not exist as a skeleton name."});
        _ -> ok
    end.

skeleton_call_pattern()->
    Name="[^()`\t\n .]+",
    Open="[(]",
    Close="[)]",
    Dot="\.",
    NullParam="[\t\n ]*",
    Param="([\t\n ]*`[^`]*`[\t\n ]*)",
    Params=Param++"(,"++Param++")*",
    ParamList="("++NullParam++"|"++Params++")",
    Name++Open++ParamList++Close++Dot.

check_skeleton_call_syntax(Call=[C|_]) when is_integer(C)->
    {ok, Mp}=re:compile(skeleton_call_pattern(),[]),
    case re:run(Call,Mp) of
        nomatch -> throw({?NITRO_ERROR, "Syntax error in skeleton."});
        _ -> ok
    end.

has_unique_skeleton_name(Name=[A|_]) when is_integer(A)->
    case length(?NITRO_HELPER:find_in_stab_by_name(Name)) of
        0 -> ok;
        _ -> throw({?NITRO_ERROR, Name++" already exists as a skeleton name."})
    end.
